import collections
import logging

import schemaflow.pipe
import schemaflow.types
import schemaflow.exceptions as _exceptions

logger = logging.getLogger('schemaflow')
logger.setLevel(logging.DEBUG)


class Pipeline(schemaflow.pipe.Pipe):
    """
    A list of :class:`~schemaflow.pipe.Pipe`'s that are applied sequentially.

    :class:`~schemaflow.pipeline.Pipeline` is a :class:`~schemaflow.pipe.Pipe` and can be part of
    another :class:`~schemaflow.pipeline.Pipeline`.

    :param pipes: a ``list`` or ``OrderedDict`` of :attr:`~schemaflow.pipe.Pipe`.
        If passed as a list, you can either pass pipes or tuples ``(name, Pipe)``.
    """
    def __init__(self, pipes):
        super().__init__()
        if isinstance(pipes, collections.OrderedDict):
            self.pipes = pipes
            return
        elif not isinstance(pipes, list):
            raise TypeError('Pipes must a list or OrderedDict')

        assert len(pipes) > 0
        #: An ``OrderedDict`` whose keys are the pipe's names or ``str(index)`` where ``index`` is the pipe's
        #: position in the sequence and the values are :attr:`~schemaflow.pipe.Pipe`'s.
        self.pipes = collections.OrderedDict()
        for i, item in enumerate(pipes):
            if isinstance(item, tuple) and len(item) == 2 and isinstance(item[1], schemaflow.pipe.Pipe) and \
                       isinstance(item[0], str):
                self.pipes[item[0]] = item[1]
            elif isinstance(item, schemaflow.pipe.Pipe):
                self.pipes[str(i)] = item
            else:
                raise TypeError('Items must be pipes or 2-element tuples of the form `(str, Pipe)`')

    @property
    def fit_requires(self):
        """
        The data schema required in :meth:`~fit`.
        """
        fit_schema = {}
        data = {}
        for key, pipe in self.pipes.items():
            # not in data => a previous pipe already added this key
            # not in transform_requires => previous pipe already required this key
            if pipe.fit_requires:
                required_data = pipe.fit_requires
            else:
                required_data = pipe.transform_requires

            new_keys = dict((key, datum_type) for key, datum_type in required_data.items()
                            if key not in data and key not in fit_schema)
            fit_schema.update(new_keys)
            data = pipe._transform_schema(data)

        return fit_schema

    @property
    def transform_requires(self):
        """
        The data schema required in :meth:`~transform`.
        """
        transform_data = {}
        data = {}
        for key, pipe in self.pipes.items():
            # not in data => a previous pipe already added this key
            # not in transform_requires => previous pipe already required this key
            new_keys = dict((key, datum_type) for key, datum_type in pipe.transform_requires.items()
                            if key not in data and key not in transform_data)
            transform_data.update(new_keys)
            data = pipe._transform_schema(data)

        return transform_data

    @property
    def transform_modifies(self):
        """
        The schema modifications that this Pipeline apply in ``transform``.

        When a key is modified more than once, changes are appended as a list.
        """
        transform_modifies = {}
        for key, pipe in self.pipes.items():
            for key_1, op in pipe.transform_modifies.items():
                if key_1 in transform_modifies:
                    if not isinstance(transform_modifies[key_1], list):
                        if op != transform_modifies[key_1]:
                            transform_modifies[key_1] = [transform_modifies[key_1], op]
                    elif op != transform_modifies[key_1][-1]:
                        transform_modifies[key_1].append(op)
                else:
                    transform_modifies.update(pipe.transform_modifies)
        return transform_modifies

    @property
    def fitted_parameters(self):
        """
        Parameters assigned to fit of each pipe.

        :return: a dictionary with the pipe's name and their respective :attr:`~schemaflow.pipe.Pipe.fitted_parameters`.
        """
        fitted_parameters = {}
        for name, pipe in self.pipes.items():
            fitted_parameters[name] = pipe.fitted_parameters
        return fitted_parameters

    @property
    def requirements(self):
        """
        Set of packages required by the Pipeline. The union of all
        :attr:`~schemaflow.pipe.Pipe.requirements` of all pipes in the Pipeline.
        """
        requirements = set()
        for pipe in self.pipes.values():
            requirements = requirements.union(pipe.requirements)
        return requirements

    def check_transform(self, data: dict=None, raise_: bool=False):
        errors = []
        for key, pipe in self.pipes.items():
            try:
                errors += pipe.check_transform(data, raise_)
            except _exceptions.SchemaFlowError as e:
                e.locations.append('of pipe \'%s\' of %s' % (key, self.__class__.__name__))
                raise e

            data = pipe._transform_schema(data)
        return errors

    def check_fit(self, data: dict, parameters: dict=None, raise_: bool=False):
        if parameters is None:
            parameters = {}

        errors = []
        for key, pipe in self.pipes.items():
            try:
                if key in parameters:
                    errors += pipe.check_fit(data, parameters[key], raise_)
                else:
                    errors += pipe.check_fit(data, {}, raise_)
            except _exceptions.SchemaFlowError as e:
                e.locations.append('of pipe \'%s\' of %s' % (key, self.__class__.__name__))
                raise e

            try:
                errors += pipe.check_transform(data, raise_)
            except _exceptions.SchemaFlowError as e:
                e.locations.append('of pipe \'%s\' of %s' % (key, self.__class__.__name__))
                raise e

            data = pipe._transform_schema(data)
        return errors

    def transform(self, data: dict):
        """
        Applies each of :meth:`~schemaflow.pipe.Pipe.transform` sequentially into ``data``.

        :param data: a dictionary of pairs ``str, object``.
        :return: the transformed data.
        """
        for key, pipe in self.pipes.items():
            data = pipe.transform(data)
        return data

    def transform_schema(self, schema: dict):
        for key, pipe in self.pipes.items():
            try:
                pipe.check_transform(schema, True)
            except _exceptions.SchemaFlowError as e:
                e.locations.append('of pipe \'%s\' of %s' % (key, self.__class__.__name__))
                raise e
            schema = pipe.transform_schema(schema)
        return schema

    def fit(self, data: dict, parameters: dict=None):
        """
        Fits the :attr:`pipes` in sequence: ``p1.fit``, ``p1.transform``, ``p2.fit``, ``p2.transform``,
        ..., ``pN.transform``.

        :param data: a dictionary of pairs ``(str, object)``.
        :param parameters: a dictionary ``{pipe_name: {str: object}}``, where each of its value is the parameters
            to be passed to the respective's pipe named ``pipe_name``.
        :return: ``None``
        """
        if parameters is None:
            parameters = {}
        for key, pipe in self.pipes.items():
            if key in parameters:
                pipe.fit(data, parameters[key])
            else:
                pipe.fit(data)
            data = pipe.transform(data)

    def _logged_transform(self, key, data):
        input_schema = schemaflow.types.infer_schema(data)
        logger.info('Started transform \'%s\' (%s): %s' % (key, self.pipes[key].__class__.__name__, input_schema))

        for error in self.pipes[key].check_transform(input_schema):
            error.locations.append('in %s' % key)
            logger.error(str(error))

        data = self.pipes[key].transform(data)
        output_schema = schemaflow.types.infer_schema(data)

        for error in self.pipes[key].check_transform_modifies(input_schema.copy(), output_schema):
            error.locations.append('in %s' % key)
            logger.error(str(error))

        logger.info('Ended   transform \'%s\' (%s): %s' % (key, self.pipes[key].__class__.__name__, output_schema))
        return data

    def logged_transform(self, data: dict):
        """
        Performs the same operation as :meth:`transform` while logging the schema on each intermediary step.

        It also logs schema inconsistencies as errors. Specifically, for each pipe, it checks if its input data
        is consistent with its :attr:`~schemaflow.pipes.Pipe.transform_requires`, and whether its output data
        is consistent with its :attr:`~schemaflow.pipes.Pipe.transform_modifies`.

        This greatly helps the Pipeline developer to identify problems in the pipeline.

        :param data: a dictionary of pairs ``str`` :class:`~schemaflow.types.Type`.
        :return: the transformed data.
        """
        for key in self.pipes:
            data = self._logged_transform(key, data)
        return data

    def logged_fit(self, data: dict, parameters: dict = None):
        """
        Performs the same operation as :meth:`fit` while logging the schema on each intermediary step.

        It also logs schema inconsistencies as errors. Specifically, for each pipe, it checks if its input data
        is consistent with its :attr:`~schemaflow.pipes.Pipe.fit_requires`, and whether its state changes is consistent
        with its :attr:`~schemaflow.pipes.Pipe.fitted_parameters`.

        This greatly helps the Pipeline developer to identify problems in the pipeline.

        :param data: a dictionary of pairs ``(str, object)``.
        :param parameters: a dictionary of pairs ``(str, object)``.
        :return: ``None``
        """
        if parameters is None:
            parameters = {}
        for key, pipe in self.pipes.items():
            schema = schemaflow.types.infer_schema(data)
            logger.info('Started fit \'%s\' (%s): %s' % (key, self.pipes[key].__class__.__name__, schema))

            if key in parameters:
                pipe.fit(data, parameters[key])
            else:
                pipe.fit(data)
            state_schema = dict((key, type(value)) for key, value in pipe.state.items())
            logger.info('Ended   fit \'%s\' (%s): state=%s' % (key, self.pipes[key].__class__.__name__, state_schema))
            data = self._logged_transform(key, data)



class SchemaFlowError(Exception):
    """
    The base exception of Pipeline
    """
    def __init__(self, locations: list=None):
        if locations is None:
            locations = []
        self.locations = locations


class NotFittedError(SchemaFlowError):
    """
    :class:`~schemaflow.exceptions.SchemaFlowError` raised when someone tries to access a non-fitted parameter.
    """
    def __init__(self, pipe, key, locations: list=None):
        self.pipe = pipe
        self.key = key
        super().__init__(locations)

    def __str__(self):
        return 'The pipe \'%s\' %s needs to be fitted before its state \'%s\' is usable' % \
            (self.pipe.__class__.__name__, ' '.join(self.locations), self.key)


class MissingRequirement(SchemaFlowError):
    """
    :class:`~schemaflow.exceptions.SchemaFlowError` raised when a requirement is missing
    """
    def __init__(self, object_type: type, requirement: str, locations: list=None):
        super().__init__(locations)
        self.object_type = object_type
        self.requirement = requirement


class WrongSchema(SchemaFlowError):
    """
    :class:`~schemaflow.exceptions.SchemaFlowError` raised when the schema of a datum is wrong (e.g. wrong shape)
    """
    def __init__(self, expected_columns, passed_columns, locations: list=None):
        super().__init__(locations)
        self.expected_columns = expected_columns
        self.passed_columns = passed_columns

    def __str__(self):
        return 'Missing arguments %s:'\
                '\nRequired arguments: %s\nPassed arguments:   %s' % \
               (' '.join(self.locations), self.expected_columns, self.passed_columns)


class WrongParameter(WrongSchema):
    """
    :class:`~schemaflow.exceptions.SchemaFlowError` raised when unexpected parameters are passed to `fit`.
    """
    def __str__(self):
        return 'Incompatible arguments %s:'\
                '\nExpected arguments: %s\nPassed arguments:   %s' % \
               (' '.join(self.locations), self.expected_columns, self.passed_columns)


class WrongType(SchemaFlowError):
    """
    :class:`~schemaflow.exceptions.SchemaFlowError` raised when the type of the datum is wrong
    """
    def __init__(self, expected_type, base_type, locations: list=None):
        super().__init__(locations)
        self.expected_type = expected_type
        self.base_type = base_type

    def __str__(self):
        return 'Wrong type %s:'\
                '\nRequired type: %s\nPassed type:   %s' % \
               (' '.join(self.locations), self.expected_type, self.base_type)


class WrongShape(SchemaFlowError):
    """
    :class:`~schemaflow.exceptions.SchemaFlowError` raised when the shape of the datum is wrong
    """
    def __init__(self, expected_shape, shape, locations: list=None):
        super().__init__(locations)
        self.expected_shape = expected_shape
        self.shape = shape

    def __str__(self):
        return 'Wrong shape %s:'\
                '\nRequired shape: %s\nPassed shape:   %s' % \
               (' '.join(self.locations), self.expected_shape, self.shape)



import unittest

import pandas as pd

from schemaflow.types import PandasDataFrame, infer_schema


class TestPandasDataFrame(unittest.TestCase):

    def test_instance_check(self):
        # ok
        type = PandasDataFrame(schema={'a': pd.np.float64, 'b': pd.np.dtype('O')})
        instance = pd.DataFrame(data={'a': [1.0, 1.0], 'b': ['s', 's']})
        self.assertEqual(type.check_schema(instance), [])

        # extra column is ok
        instance = pd.DataFrame(data={'a': [1.0, 1.0], 'b': ['s', 's'], 'c': [1, 1]})
        self.assertEqual(type.check_schema(instance), [])

        # missing column
        instance = pd.DataFrame(data={'a': [1.0, 1.0]})
        self.assertEqual(len(type.check_schema(instance)), 1)

        # wrong column type
        instance = pd.DataFrame(data={'a': [1.0, 1.0], 'b': [1.0, 1.0]})
        self.assertEqual(len(type.check_schema(instance)), 1)

        # wrong datum type
        instance = 1
        self.assertEqual(len(type.check_schema(instance)), 1)

    def test_type_check(self):
        # ok
        type = PandasDataFrame(schema={'a': pd.np.float64, 'b': pd.np.dtype('O')})
        self.assertEqual(type.check_schema(type), [])

        # missing column
        instance = pd.DataFrame(data={'a': [1.0, 1.0]})
        self.assertEqual(len(type.check_schema(PandasDataFrame(schema={'a': pd.np.float64}))), 1)

        # wrong column type
        self.assertEqual(len(type.check_schema(PandasDataFrame(schema={'a': pd.np.float64, 'b': pd.np.float64}))), 1)

        # wrong datum type
        self.assertEqual(len(type.check_schema(int)), 1)

    def test_infer(self):
        instance = pd.DataFrame(data={'a': [1.0, 1.0], 'b': ['a', 'b']})

        schema = infer_schema({'a': instance})
        self.assertEqual(schema, {'a': PandasDataFrame(schema={'a': pd.np.float64, 'b': pd.np.dtype('O')})})



import unittest
import numpy as np

from schemaflow.types import Array, infer_schema
from schemaflow import exceptions


class TestArray(unittest.TestCase):

    def test_instance_check(self):
        # 2D with N features
        instance = np.array([[1.0], [2.0]])
        array_type = Array(float, shape=(None, None))
        self.assertEqual(array_type.check_schema(instance), [])

        # 2D with 1 feature
        instance = np.array([[1.0], [2.0]])
        array_type = Array(float, shape=(None, 1))
        self.assertEqual(array_type.check_schema(instance), [])

        # 2D with 2 features != 2D with 1 feature
        instance = np.array([[1.0], [2.0]])
        array_type = Array(float, shape=(None, 2))
        self.assertEqual(len(array_type.check_schema(instance)), 1)

        # 1D != 2D
        instance = np.array([[1.0], [2.0]])
        array_type = Array(float, shape=(None,))
        self.assertEqual(len(array_type.check_schema(instance)), 1)

        # 1D == 1D
        instance = np.array([1.0, 2.0])
        array_type = Array(float, shape=(None,))
        self.assertEqual(array_type.check_schema(instance), [])

        instance = np.float64(2)
        array_type = Array(float, shape=(None,))
        self.assertEqual(len(array_type.check_schema(instance)), 1)

        instance = np.array([object, 2.0])
        array_type = Array(float, shape=(None,))
        self.assertEqual(len(array_type.check_schema(instance)), 1)

    def test_type_check(self):
        # 2D with N features
        array_type = Array(float, shape=(None, None))
        self.assertEqual(array_type.check_schema(array_type), [])

        # 2D with 1 feature
        instance = Array(float, shape=(None, 1))
        self.assertEqual(array_type.check_schema(instance), [])

        # 2D with 2 features != 2D with 1 feature
        instance = Array(float, shape=(1, 1))
        array_type = Array(float, shape=(None, 2))
        self.assertEqual(len(array_type.check_schema(instance)), 1)

        # 1D != 2D
        instance = Array(float, shape=(1, 1))
        array_type = Array(float, shape=(None,))
        self.assertEqual(len(array_type.check_schema(instance)), 1)

        # 1D == 1D
        array_type = Array(float, shape=(None,))
        self.assertEqual(array_type.check_schema(array_type), [])

        instance = np.float64
        array_type = Array(float, shape=(None,))
        self.assertEqual(len(array_type.check_schema(instance)), 1)

        instance = np.array([object, 2.0]).dtype
        array_type = Array(float, shape=(None,))
        self.assertEqual(len(array_type.check_schema(instance)), 1)

        instance = Array(float, shape=(None, 1))
        array_type = Array(np.float64, shape=(1,))
        with self.assertRaises(exceptions.WrongShape) as e:
            array_type.check_schema(instance, True)
        self.assertIn('Wrong shape', str(e.exception))

    def test_infer(self):
        instance = np.array([[1.0], [2.0]])

        schema = infer_schema({'a': instance})
        self.assertEqual(schema, {'a': Array(np.float64, (2, 1))})



import unittest
import datetime

import numpy as np
import pyspark
from pyspark.sql.types import Row

from schemaflow.types import PySparkDataFrame, infer_schema


class PySparkTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        conf = pyspark.SparkConf().setMaster('local[1]').setAppName("testing")
        cls.sc = pyspark.SparkContext(conf=conf)
        cls.sqlContext = pyspark.SQLContext(cls.sc)

    @classmethod
    def tearDownClass(cls):
        cls.sc.stop()


class TestPySparkDataFrame(PySparkTestCase):

    def test_type_check(self):
        # ok
        type = PySparkDataFrame(schema={'a': float, 'b': np.dtype('O')})
        instance = self.sqlContext.createDataFrame(data=[Row(a=1.0, b='s'), Row(a=1.0, b='s')])
        self.assertEqual(type.check_schema(instance), [])

        # extra column is ok
        instance = self.sqlContext.createDataFrame(data=[Row(a=1.0, b='s', c=1.0),
                                                         Row(a=1.0, b='s', c=1.0)])
        self.assertEqual(type.check_schema(instance), [])

        # missing column
        instance = self.sqlContext.createDataFrame(data=[Row(a=1.0), Row(a=1.0)])
        self.assertEqual(len(type.check_schema(instance)), 1)

        # wrong column type
        instance = self.sqlContext.createDataFrame(data=[Row(a=1.0, b=1),
                                                         Row(a=1.0, b=1)])
        self.assertEqual(len(type.check_schema(instance)), 1)

    def test_date_time(self):
        instance_type = PySparkDataFrame(schema={'a': datetime.datetime})
        instance = self.sqlContext.createDataFrame(data=[Row(a=datetime.datetime.now())])
        self.assertEqual(len(instance_type.check_schema(instance)), 0)

        instance_type = PySparkDataFrame(schema={'a': datetime.date})
        instance = self.sqlContext.createDataFrame(data=[Row(a=datetime.datetime.now().date())])
        self.assertEqual(len(instance_type.check_schema(instance)), 0)

        # wrong types
        instance_type = PySparkDataFrame(schema={'a': datetime.datetime})
        instance = self.sqlContext.createDataFrame(data=[Row(a=datetime.datetime.now().date())])
        self.assertEqual(len(instance_type.check_schema(instance)), 1)

        instance_type = PySparkDataFrame(schema={'a': datetime.date})
        instance = self.sqlContext.createDataFrame(data=[Row(a=datetime.datetime.now())])
        self.assertEqual(len(instance_type.check_schema(instance)), 1)

    def test_int_float(self):
        instance_type = PySparkDataFrame(schema={'a': float})
        instance = self.sqlContext.createDataFrame(data=[Row(a=1.0)])
        self.assertEqual(len(instance_type.check_schema(instance)), 0)

        instance_type = PySparkDataFrame(schema={'a': int})
        instance = self.sqlContext.createDataFrame(data=[Row(a=1)])
        self.assertEqual(len(instance_type.check_schema(instance)), 0)

        # wrong types
        instance_type = PySparkDataFrame(schema={'a': float})
        instance = self.sqlContext.createDataFrame(data=[Row(a=1)])
        self.assertEqual(len(instance_type.check_schema(instance)), 1)

        instance_type = PySparkDataFrame(schema={'a': int})
        instance = self.sqlContext.createDataFrame(data=[Row(a=1.0)])
        self.assertEqual(len(instance_type.check_schema(instance)), 1)

    def test_bool(self):
        instance_type = PySparkDataFrame(schema={'a': bool})
        instance = self.sqlContext.createDataFrame(data=[Row(a=False)])
        self.assertEqual(len(instance_type.check_schema(instance)), 0)

        instance = self.sqlContext.createDataFrame(data=[Row(a=1)])
        self.assertEqual(len(instance_type.check_schema(instance)), 1)

    def test_infer(self):
        instance = self.sqlContext.createDataFrame(data=[Row(a=1.0, b='a'), Row(a=1.0, b='b')])

        schema = infer_schema({'a': instance})
        self.assertEqual(schema, {'a': PySparkDataFrame(schema={'a': float, 'b': np.dtype('O')})})



import unittest

import numpy as np
import pandas as pd

from schemaflow.pipeline import Pipeline
from schemaflow.pipe import Pipe
from schemaflow import types, ops


class Pipe1(Pipe):
    transform_requires = {
        'x': types.PandasDataFrame(schema={'a': np.float64, 'b': np.float64}),
    }

    transform_modifies = {
        'x': ops.ModifyDataFrame({'a * b': ops.Set(np.float64)})
    }

    def transform(self, data: dict):
        data['x']['a * b'] = data['x']['a'] * data['x']['b']
        return data


class Pipe2(Pipe):
    transform_requires = {
        'x': types.PandasDataFrame(schema={'a': np.float64, 'b': np.float64}),
    }

    transform_modifies = {
        'x': ops.ModifyDataFrame({'a': ops.Drop()})
    }

    def transform(self, data: dict):
        data['x'] = data['x'].drop('a', axis=1)
        return data


class TestPipeline(unittest.TestCase):

    def test_set(self):
        p = Pipeline([Pipe1()])

        result = p.transform({'x': pd.DataFrame({'a': [2.0], 'b': [2.0]})})
        self.assertEqual(result['x'].loc[:, 'a * b'].values, [4.0])

        self.assertEqual(p.transform_modifies, Pipe1.transform_modifies)

    def test_drop(self):
        p = Pipeline([Pipe2()])

        result = p.transform({'x': pd.DataFrame({'a': [2.0], 'b': [2.0]})})
        self.assertEqual(len(result['x'].columns), 1)

        self.assertEqual(p.transform_modifies, Pipe2.transform_modifies)

    def test_combine(self):
        p = Pipeline([Pipe1(), Pipe2()])

        result = p.transform({'x': pd.DataFrame({'a': [2.0], 'b': [2.0]})})
        self.assertEqual(len(result['x'].columns), 2)

        self.assertEqual(p.transform_modifies,
                         {'x': [Pipe1.transform_modifies['x'], Pipe2.transform_modifies['x']]})

        schema = p.transform_schema({'x': types.PandasDataFrame({'a': np.float64, 'b': np.float64})})

        self.assertEqual(schema['x'], types.PandasDataFrame({'b': np.float64,
                                                             'a * b': np.float64}))



import unittest

from schemaflow.types import List, Tuple, infer_schema


class TestListTuple(unittest.TestCase):

    def test_instance_check(self):
        instance = [1.0, 2.0]
        array_type = List(float)
        self.assertEqual(array_type.check_schema(instance), [])

        instance = [1.0, 'a']
        array_type = List(float)
        self.assertEqual(len(array_type.check_schema(instance)), 1)

        instance = (1.0, 2.0)
        array_type = List(float)
        self.assertEqual(len(array_type.check_schema(instance)), 1)

        instance = (1.0, 2.0)
        array_type = Tuple(float)
        self.assertEqual(array_type.check_schema(instance), [])

    def test_type_check(self):
        array_type = List(float)
        self.assertEqual(array_type.check_schema(List(float)), [])

        array_type = List(float)
        self.assertEqual(len(array_type.check_schema(List(object))), 1)

        array_type = List(float)
        self.assertEqual(len(array_type.check_schema(Tuple(float))), 1)

        array_type = Tuple(float)
        self.assertEqual(array_type.check_schema(Tuple(float)), [])

    def test_infer(self):
        instance = [1.0, 2.0]

        schema = infer_schema({'a': instance})
        self.assertEqual(schema, {'a': List(float)})



import datetime
import importlib.util

import schemaflow.exceptions as _exceptions


def _requirement_fulfilled(requirement: str):
    """
    Returns whether a requirement is fulfilled.

    :return: bool
    """
    return importlib.util.find_spec(requirement) is not None


def _all_subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in _all_subclasses(c)])


def infer_schema(data: dict):
    subclasses = list(subclass for subclass in _all_subclasses(Type)
                      if subclass.requirements_fulfilled and not subclass.__name__.startswith('_'))

    schema = {}
    for key, value in data.items():
        for subclass in subclasses:
            if isinstance(value, subclass.base_type()):
                schema[key] = subclass.infer(value)
                break
        else:
            schema[key] = type(value)
    return schema


def _get_type(instance_type):
    if not isinstance(instance_type, Type):
        instance_type = _LiteralType(instance_type)
    return instance_type


class Type:
    """
    The base type of all types. Used to declare new types to be used in :class:`schemaflow.pipe.Pipe`.

    The class attribute :attr:`requirements` (a set of strings) is used to define if using this type has
    package requirements (e.g. `numpy`).
    """
    requirements = {}  #: set of packages required for this type to be usable.

    @classmethod
    def base_type(cls):
        """
        A class property that returns the underlying type of this Type.
        :return:
        """
        raise NotImplementedError

    @classmethod
    def infer(cls, instance):
        raise NotImplementedError

    @classmethod
    def requirements_fulfilled(cls):
        """
        Returns whether this Type has its requirements fulfilled.

        :return: bool
        """
        return all(_requirement_fulfilled(requirement) for requirement in cls.requirements)

    def _check_as_instance(self, instance: object, raise_: bool):
        raise NotImplementedError

    def _check_as_type(self, instance, raise_: bool):
        if type(instance) != type(self):
            exception = _exceptions.WrongType(instance, self)
            if raise_:
                raise exception
            return [exception]
        return []

    def check_schema(self, instance: object, raise_: bool=False):
        """
        Checks that the instance has the correct type and schema (composite types).

        :param instance: a datum in either its representation form or on its schema form.
        :param raise_:
        :return: a list of exceptions
        """
        if isinstance(instance, Type):
            return self._check_as_type(instance, raise_)
        else:
            return self._check_as_instance(instance, raise_)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__


class _LiteralType(Type):
    """
    A :class:`Type` that wraps literal types (e.g. ``float``). Used internally only.
    """
    def __init__(self, base_type):
        assert not isinstance(base_type, Type)
        assert not isinstance(base_type, (list, tuple))
        self._base_type = base_type

    def __repr__(self):
        return 'L(%s)' % self._base_type

    @property
    def base_type(self):
        return self._base_type

    def _check_as_instance(self, instance: object, raise_: bool):
        if not isinstance(instance, self._base_type):
            exception = _exceptions.WrongType(self._base_type, type(instance))
            if raise_:
                raise exception
            return [exception]
        return []


class _DataFrame(Type):
    """
    Abstract schemaflow representation of a DataFrame. See subclasses for Pandas and PySpark.
    """
    def __init__(self, schema: dict):
        """
        :param schema: dictionary of `(column_name, type)`.
        """
        self.schema = schema.copy()
        for column, base_type in self.schema.items():
            if not isinstance(base_type, Type):
                self.schema[column] = _LiteralType(base_type)

    def __getitem__(self, key):
        return self.schema[key]

    def __setitem__(self, key, value):
        self.schema[key] = value

    def __delitem__(self, key):
        del self.schema[key]

    @classmethod
    def infer(cls, instance):
        assert isinstance(instance, cls.base_type())
        return cls(schema=cls._get_schema(instance))

    @staticmethod
    def _get_schema(instance):
        """
        Return the DataFrame's schema from an instance
        """
        raise NotImplementedError

    def _check_schema(self, schema, raise_: bool):
        exceptions = []
        for column in self.schema:
            if column not in schema:
                exception = _exceptions.WrongSchema(column, set(schema.keys()))
                if raise_:
                    raise exception
                exceptions.append(exception)
            else:
                column_type = schema[column].base_type
                expected_type = self.schema[column].base_type
                if expected_type != column_type:
                    exception = _exceptions.WrongType(
                        expected_type, column_type, 'column \'%s\'' % column)
                    if raise_:
                        raise exception
                    exceptions.append(exception)
        return exceptions

    def _check_as_type(self, instance, raise_: bool):
        exceptions = super()._check_as_type(instance, raise_)
        if not exceptions:
            exceptions += self._check_schema(instance.schema, raise_)
        return exceptions

    def _check_as_instance(self, instance: object, raise_: bool):
        if not isinstance(instance, self.base_type()):
            exception = _exceptions.WrongType(self.base_type(), type(instance))
            if raise_:
                raise exception
            return [exception]
        return self._check_schema(self._get_schema(instance), raise_)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.schema)


class PySparkDataFrame(_DataFrame):
    """
    Representation of a pyspark.sql.DataFrame. Requires ``pyspark``.
    """
    requirements = {'pyspark'}

    @classmethod
    def base_type(cls):
        import pyspark.sql
        return pyspark.sql.DataFrame

    @staticmethod
    def _get_schema(instance):
        import pyspark.sql.types
        import numpy
        mapping = {
            pyspark.sql.types.LongType: _LiteralType(int),
            pyspark.sql.types.DoubleType: _LiteralType(float),
            pyspark.sql.types.BooleanType: _LiteralType(bool),
            pyspark.sql.types.StringType: _LiteralType(numpy.dtype('O')),
            pyspark.sql.types.DateType: _LiteralType(datetime.date),
            pyspark.sql.types.TimestampType: _LiteralType(datetime.datetime),
        }
        return dict((f.name, mapping[type(f.dataType)]) for f in instance.schema.fields)


class PandasDataFrame(_DataFrame):
    """
    Representation of a pandas.DataFrame. Requires ``pandas``.
    """
    requirements = {'pandas'}

    @classmethod
    def base_type(cls):
        import pandas
        return pandas.DataFrame

    @staticmethod
    def _get_schema(instance):
        return dict((column, _get_type(column_dtype)) for column, column_dtype in instance.dtypes.items())


class _Container(Type):

    @classmethod
    def base_type(cls):
        return list

    def __init__(self, items_type):
        if not isinstance(items_type, Type):
            items_type = _LiteralType(items_type)
        self._items_type = items_type

    @classmethod
    def infer(cls, instance):
        assert isinstance(instance, cls.base_type())

        inferred_items_type = object
        if len(instance):
            inferred_items_type = type(instance[0])
            if any(type(item) != inferred_items_type for item in instance):
                inferred_items_type = object

        return cls(inferred_items_type)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self._items_type.base_type.__name__)

    def _check_as_type(self, instance, raise_: bool):
        exceptions = super()._check_as_type(instance, raise_)
        if not exceptions and self._items_type.base_type != instance._items_type.base_type:
            exception = _exceptions.WrongType(self, instance)
            if raise_:
                raise exception
            exceptions += [exception]
        return exceptions

    def _check_as_instance(self, instance: object, raise_: bool):
        if not isinstance(instance, self.base_type()):
            exception = _exceptions.WrongType(self.base_type(), type(instance))
            if raise_:
                raise exception
            return [exception]

        exceptions = []
        if not exceptions:
            for i, item in enumerate(instance):
                exceptions += self._items_type.check_schema(item, raise_)
        return exceptions


class List(_Container):
    pass


class Tuple(_Container):
    @classmethod
    def base_type(cls):
        return tuple


class Array(_Container):
    """
    Representation of a numpy.array. Requires ``numpy``.
    """
    requirements = {'numpy'}

    def __init__(self, items_type: type, shape=None):
        import numpy
        assert isinstance(items_type, numpy.dtype) or issubclass(items_type, (numpy.generic, float, int, bool))
        if items_type == float:
            items_type = numpy.float64
        super().__init__(_LiteralType(items_type))
        assert isinstance(shape, (type(None), tuple))
        self.shape = shape

    def __repr__(self):
        return '%s(%s, %s)' % (self.__class__.__name__, self._items_type.base_type, self.shape)

    @classmethod
    def infer(cls, instance):
        assert isinstance(instance, cls.base_type())

        return cls(instance.dtype, instance.shape)

    def _check_as_type(self, instance, raise_: bool):
        exceptions = super()._check_as_type(instance, raise_)
        if not exceptions:
            if not self._is_valid_shape(instance.shape):
                exception = _exceptions.WrongShape(self.shape, instance.shape)
                if raise_:
                    raise exception
                exceptions.append(exception)
        return exceptions

    def _check_as_instance(self, instance: object, raise_: bool):
        if not isinstance(instance, self.base_type()):
            exception = _exceptions.WrongType(self.base_type(), type(instance))
            if raise_:
                raise exception
            return [exception]

        exceptions = []
        assert hasattr(instance, 'dtype')
        if instance.dtype == self._items_type.base_type and hasattr(instance, 'shape'):
            if not self._is_valid_shape(instance.shape):
                exception = _exceptions.WrongShape(self.shape, instance.shape)
                if raise_:
                    raise exception
                exceptions.append(exception)
        else:
            exception = _exceptions.WrongType(self._items_type.base_type, instance.dtype)
            if raise_:
                raise exception
            exceptions.append(exception)
        return exceptions

    def _is_valid_shape(self, shape):
        if self.shape is not None:
            if len(shape) != len(self.shape):
                return False
            for required_d, d in zip(self.shape, shape):
                if required_d is not None and d != required_d:
                    return False
        return True

    @classmethod
    def base_type(cls):
        import numpy
        return numpy.ndarray



import unittest
import logging
import collections

from schemaflow.pipeline import Pipeline
from schemaflow.pipe import Pipe
from schemaflow import types
from schemaflow import exceptions


class Pipe1(Pipe):
    requirements = {'a1'}

    transform_requires = {
        'x': types.List(str),
    }

    transform_modifies = {
        'x': types.List(float),
    }

    def transform(self, data: dict):
        data['x'] = [float(x_i) for x_i in data['x']]
        return data


class Pipe2(Pipe):
    requirements = {'a2'}

    transform_requires = {
        'x': types.List(float),
    }

    fit_requires = {
        'x': types.List(float),
    }

    fitted_parameters = {'mean': float, 'var': float}

    fit_parameters = {'unused': float}

    transform_modifies = {
        'x': types.List(float),
    }

    def fit(self, data: dict, parameters: dict=None):
        self['mean'] = sum(data['x']) / len(data['x'])
        self['var'] = sum([x_i**2 for x_i in data['x']]) / len(data['x']) - self['mean']**2

    def transform(self, data: dict):
        data['x'] = [(x_i - self['mean'])/self['var']**0.5 for x_i in data['x']]
        return data


class Pipe3(Pipe):
    transform_requires = {
        'x1': types.List(str),
        'x': types.List(float),
    }


class Pipe4(Pipe):
    fit_requires = {
        'x1': types.List(float),
    }

    fitted_parameters = {
        'mean': float
    }

    transform_requires = {
        'x': types.List(float),
    }


class PipeWrongTransform(Pipe):
    """
    Claims to convert to float, but converts to int
    """
    requirements = {'a1'}

    transform_requires = {
        'x': types.List(str),
    }

    transform_modifies = {
        'x': types.List(float),
    }

    def transform(self, data: dict):
        data['x'] = [int(x_i) for x_i in data['x']]
        return data


class TestPipeline(unittest.TestCase):

    def test_check_fit(self):
        p = Pipeline([Pipe1(), Pipe2()])

        # ok
        self.assertEqual(p.check_fit({'x': ['1']}, {'1': {'unused': 1.0}}), [])
        # ok with raise
        self.assertEqual(p.check_fit({'x': ['1']}, {'1': {'unused': 1.0}}, True), [])

        # not ok
        self.assertEqual(len(p.check_fit({'x': 1}, {'1': {'unused': 1.0}})), 1)

        # not ok with raise
        with self.assertRaises(exceptions.WrongType):
            p.check_fit({'x': [1]}, {'1': {'unused': 1.0}}, True)

        p = Pipeline([Pipe2()])
        with self.assertRaises(exceptions.WrongType):
            p.check_fit({'x': ['1']}, {'1': {'unused': 1.0}}, True)

    def test_check_transform(self):
        p = Pipeline([Pipe1(), Pipe2()])

        # ok
        self.assertEqual(p.check_transform({'x': ['1']}), [])
        # ok with raise
        self.assertEqual(p.check_transform({'x': ['1']}, True), [])

        # not ok
        self.assertEqual(len(p.check_transform({'x': 1})), 1)

        # not ok with raise
        with self.assertRaises(exceptions.WrongType):
            p.check_transform({'x': [1]}, True)

    def test_requirements(self):
        p = Pipeline([Pipe1(), Pipe2()])
        self.assertEqual(p.requirements, {'a1', 'a2'})

        self.assertEqual(len(p.check_requirements), 2)

    def test_basic(self):
        p = Pipeline([Pipe1(), Pipe2()])

        self.assertEqual(p.fitted_parameters, {'0': {}, '1': {'mean': float, 'var': float}})
        self.assertEqual(p.transform_requires, {'x': types.List(str)})
        self.assertEqual(p.fit_requires, {'x': types.List(str)})

        p.fit({'x': ['1', '2', '3']}, {'1': {'unused': 1.0}})
        result = p.transform({'x': ['1', '2', '3']})

        # std([1,2,3]) == 0.816496580927726
        self.assertEqual(result['x'], [-1.2247448713915887, 0.0, 1.2247448713915887])

    def test_custom_parameters(self):
        p = Pipeline([('1', Pipe1()), ('2', Pipe2())])

        p.fit({'x': ['1', '2', '3']}, {'2': {'unused': 1.0}})
        result = p.transform({'x': ['1', '2', '3']})

        # std([1,2,3]) == 0.816496580927726
        self.assertEqual(result['x'], [-1.2247448713915887, 0.0, 1.2247448713915887])

    def test_custom_init(self):
        pipes = collections.OrderedDict([('1', Pipe1()), ('2', Pipe2())])
        p = Pipeline(pipes)

        self.assertEqual(p.pipes, pipes)

        with self.assertRaises(TypeError):
            Pipeline([('1', 1)])

        with self.assertRaises(TypeError):
            Pipeline(Pipe1())

    def test_two_transform_data(self):
        # P1 needs 'x', P2 needs 'x1'
        p = Pipeline([Pipe1(), Pipe3(), Pipe2()])

        self.assertEqual(p.transform_requires, {'x': types.List(str), 'x1': types.List(str)})

        self.assertEqual(p.fit_requires, {'x': types.List(str), 'x1': types.List(str)})

        self.assertEqual(p.transform_modifies, {'x': types.List(float)})

    def test_transform_schema(self):
        # P1 needs 'x', P2 needs 'x1'
        p = Pipeline([Pipe1(), Pipe3(), Pipe2()])

        # 'x1' is passed along without modification
        self.assertEqual(p.transform_schema({'x': types.List(str), 'x1': types.List(str)}),
                         {'x': types.List(float), 'x1': types.List(str)})

        with self.assertRaises(exceptions.WrongSchema) as e:
            p.transform_schema({'y': types.List(str)})
        self.assertIn('in transform of pipe \'0\' of Pipeline', str(e.exception))

        with self.assertRaises(exceptions.WrongSchema) as e:
            p.transform_schema({'x': types.List(str)})
        self.assertIn('in transform of pipe \'1\' of Pipeline', str(e.exception))

    def test_two_fit_schema(self):
        # P4 fit-needs 'x1', P2 fit-needs 'x' (float) => fit_requires needs both on its first type-occurrence
        p = Pipeline([Pipe1(), Pipe4(), Pipe2()])

        self.assertEqual(p.transform_requires, {'x': types.List(str)})

        self.assertEqual(p.fit_requires, {'x': types.List(str), 'x1': types.List(float)})

        self.assertEqual(p.transform_modifies, {'x': types.List(float)})


class MockLoggingHandler(logging.Handler):
    """Mock logging handler to check for expected logs."""
    # see https://stackoverflow.com/a/1049375/931303

    def __init__(self, *args, **kwargs):
        self.messages = {
            'debug': [],
            'info': [],
            'warning': [],
            'error': [],
            'critical': [],
        }
        logging.Handler.__init__(self, *args, **kwargs)

    def reset(self):
        self.messages = {
            'debug': [],
            'info': [],
            'warning': [],
            'error': [],
            'critical': [],
        }

    def emit(self, record):
        self.messages[record.levelname.lower()].append(record.getMessage())


class TestPipelineLogging(unittest.TestCase):

    def setUp(self):
        logger = logging.getLogger('schemaflow')
        self._handler = MockLoggingHandler()
        logger.addHandler(self._handler)

    def tearDown(self):
        logging.getLogger().removeHandler(self._handler)

    def test_logged(self):
        p = Pipeline([Pipe1(), Pipe2()])

        p.logged_fit({'x': ['1', '2', '3']}, {'1': {'unused': 1.0}})
        self.assertEqual(self._handler.messages['error'], [])
        self.assertEqual(len(self._handler.messages['info']), 8)

        result = p.logged_transform({'x': ['1', '2', '3']})
        self.assertEqual(self._handler.messages['error'], [])
        self.assertEqual(len(self._handler.messages['info']), 8 + 4)

        # std([1,2,3]) == 0.816496580927726
        self.assertEqual(result['x'], [-1.2247448713915887, 0.0, 1.2247448713915887])

    def test_logged_transform(self):
        p = Pipeline([Pipe1(), Pipe2()])
        p.fit({'x': ['1', '2', '3']}, {'1': {'unused': 1.0}})
        p.logged_transform({'x': ['1', '2', '3']})
        self.assertEqual(self._handler.messages['error'], [])
        self.assertEqual(len(self._handler.messages['info']), 4)

        self._handler.reset()

        p.logged_transform({'x': [1, 2, 3]})
        self.assertEqual(len(self._handler.messages['error']), 1)
        self.assertEqual(len(self._handler.messages['info']), 4)

        self._handler.reset()

        p = Pipeline([PipeWrongTransform(), Pipe2()])
        p.fit({'x': ['1', '2', '3']}, {'1': {'unused': 1.0}})
        p.logged_transform({'x': ['1', '2', '3']})
        # 2 errors: 1 transform_modify and 1 for wrong input to second pipe
        self.assertEqual(len(self._handler.messages['error']), 2)
        self.assertEqual(self._handler.messages['error'][0],
                         "Wrong type in result 'x' of modified data from transform in 0:\nRequired type: List(float)\nPassed type:   List(int)")
        self.assertEqual(self._handler.messages['error'][1],
                         "Wrong type in argument 'x' of transform in 1:\nRequired type: List(float)\nPassed type:   List(int)")
        self.assertEqual(len(self._handler.messages['info']), 4)



import schemaflow.types
import schemaflow.ops
import schemaflow.exceptions as _exceptions


def _check_schema_keys(schema: dict, expected: dict, error_location: list, raise_: bool=False):
    schema_keys = set(schema.keys())
    required_keys = set(expected.keys())

    exceptions = []
    if len(required_keys - schema_keys):
        exception = _exceptions.WrongSchema(required_keys, schema_keys, error_location)
        if raise_:
            raise exception
        exceptions.append(exception)
    return exceptions


def _check_schema_types(schema: dict, expected: dict, error_location: str, raise_: bool=False):
    exceptions = []
    for key, expected_type in expected.items():
        if key in schema:
            try:
                new_exceptions = expected_type.check_schema(schema[key], raise_)
            except _exceptions.SchemaFlowError as e:
                e.locations.append(error_location % key)
                raise e
            for exception in new_exceptions:
                exception.locations.append(error_location % key)
            exceptions += new_exceptions
    return exceptions


class Pipe:
    """
    A Pipe represents a stateful data transformation.

    Data in this context consists of a Python dictionary whose each value is a type with some representation of data,
    either in-memory (e.g. ``float``, ``pandas.DataFrame``) of remote (e.g. ``pyspark.sql.DataFrame``, ``sqlalchemy``).

    A :class:`Pipe` is defined by:

    - a method :meth:`transform` that:

        - uses the keys :attr:`transform_modifies` from ``data``
        - uses the :attr:`state`
        - modifies the keys in :attr:`transform_modifies` in ``data``

    - a method :meth:`fit` that:

        - uses (training) keys :attr:`fit_requires` from ``data``
        - uses (passed) :attr:`fit_parameters`
        - modifies the keys :attr:`fitted_parameters` in :attr:`state`

    - a set of :attr:`requirements` (a set of package names, e.g. ``{'pandas'}``) of the transformation

    All :attr:`transform_modifies` and :attr:`fit_requires` have a :class:`~schemaflow.types.Type` that can
    be used to check that the Pipe's input is consistent, with

        - :meth:`check_fit`
        - :meth:`check_transform`

    The existence of the requirements can be checked using

        - :meth:`check_requirements`

    The rational is that you can run ``check_*`` with access only to the data's schema.
    This is specially important when the schemaflow is an expensive operation.
    """
    #: set of packages required by the Pipe.
    requirements = set()

    #: the data schema required in :meth:`~fit`;
    #: a dictionary ``str``: :class:`~schemaflow.types.Type`.
    fit_requires = {}

    #: the data schema required in :meth:`~transform`;
    #: a dictionary ``str``: :class:`~schemaflow.type.Type`.
    transform_requires = {}

    #: parameters' schema passed to :meth:`~fit`
    fit_parameters = {}

    #: schema of the parameters assigned in :meth:`~fit`
    fitted_parameters = {}

    #: type and key of :meth:`~transform`
    transform_modifies = {}

    def __init__(self):
        self.state = {}  #: A dictionary with the states of the Pipe. Use [] operator to access and modify it.

    def __setitem__(self, key, value):
        self.state.__setitem__(key, value)

    def __getitem__(self, key):
        if key not in self.state:
            raise _exceptions.NotFittedError(self, key)
        return self.state.__getitem__(key)

    @property
    def check_requirements(self):
        """
        Checks for requirements.

        :return: a list of exceptions with missing requirements
        """
        exceptions = []
        for requirement in self.requirements:
            if not schemaflow.types._requirement_fulfilled(requirement):
                exceptions.append(_exceptions.MissingRequirement(self.__class__, requirement))

        all_types = list(self.fit_requires.values()) + list(self.transform_requires.values()) + \
                    list(self.fit_parameters.values()) + list(self.fitted_parameters.values())

        for value_type in all_types:
            if isinstance(value_type, schemaflow.types.Type):
                for requirement in value_type.requirements:
                    if not schemaflow.types._requirement_fulfilled(requirement):
                        exceptions.append(_exceptions.MissingRequirement(value_type.__class__, requirement))

        return exceptions

    def check_transform_modifies(self, input_schema: dict, output_schema: dict):
        expected_schema = self._transform_schema(input_schema.copy())
        expected_schema = {key: schemaflow.types._get_type(value) for key, value in expected_schema.items()}

        exceptions = _check_schema_keys(output_schema, expected_schema, ['in modified data from transform'])

        exceptions += _check_schema_types(output_schema, expected_schema, 'in result \'%s\' of modified data from transform')

        return exceptions

    def check_fit(self, data: dict, parameters: dict=None, raise_: bool=False):
        """
        Checks that a given data has a valid schema to be used in :meth:`fit`.

        :param data: a dictionary with either ``(str, Type)`` or ``(str, instance)``
        :param parameters: a dictionary with either ``(str, Type)`` or ``(str, instance)``
        :param raise_: whether it should raise the first found exception or list them all (default: list them)
        :return: a list of (subclasses of) :class:`~pipeline.exceptions.PipelineError` with all failed checks.
        """
        if parameters is None:
            parameters = {}

        exceptions = _check_schema_keys(data, self.fit_requires, ['in fit'], raise_)

        expected_schema = {key: schemaflow.types._get_type(value) for key, value in self.fit_requires.items()}
        exceptions += _check_schema_types(data, expected_schema, 'in argument \'%s\' of fit', raise_)

        # check that parameters are correct
        parameters_keys = set(parameters.keys())
        expected_parameters = set(self.fit_parameters.keys())
        if parameters_keys != expected_parameters:
            exception = _exceptions.WrongParameter(
                expected_parameters,
                parameters_keys, ['in fit'])
            if raise_:
                raise exception
            exceptions.append(exception)

        expected_schema = {key: schemaflow.types._get_type(value) for key, value in self.fit_parameters.items()}
        exceptions += _check_schema_types(parameters, expected_schema, 'in parameter \'%s\' of fit', raise_)

        return exceptions

    def check_transform(self, data: dict, raise_: bool=False):
        """
        Checks that a given data has a valid schema to be used in :meth:`transform`.

        :param data: a dictionary with either ``(str, Type)`` or ``(str, instance)``
        :param raise_: whether it should raise the first found exception or list them all (default: list them)
        :return: a list of (subclasses of) :class:`schemaflow.exceptions.SchemaFlowError` with all missing arguments.
        """
        exceptions = _check_schema_keys(data, self.transform_requires, ['in transform'], raise_)

        expected_schema = {key: schemaflow.types._get_type(value) for key, value in self.transform_requires.items()}
        exceptions += _check_schema_types(data, expected_schema, 'in argument \'%s\' of transform', raise_)
        return exceptions

    def _transform_schema(self, schema: dict):
        for key, value in self.transform_modifies.items():
            if not isinstance(value, list):
                value = [value]
            for transformation in value:
                if isinstance(transformation, schemaflow.ops.Operation):
                    schema = transformation.transform(key, schema)
                else:
                    schema[key] = transformation
        return schema

    def transform_schema(self, schema: dict):
        """
        Transforms the ``schema`` into a new schema based on :attr:`~transform_modifies`.

        :param schema: a dictionary of pairs ``str`` :class:`~schemaflow.types.Type`.
        :return: the new schema.
        """
        self.check_transform(schema, True)
        return self._transform_schema(schema)

    def fit(self, data: dict, parameters: dict=None):
        """
        Modifies the instance's :attr:`state`.

        :param data: a dictionary of pairs ``(str, object)``.
        :param parameters: a dictionary of pairs ``(str, object)``.
        :return: ``None``
        """

    def transform(self, data: dict):
        """
        Modifies the data keys identified in :attr:`transform_modifies`.

        :param data: a dictionary of pairs ``(str, object)``.
        :return: the modified data
        """
        return data



import os
import setuptools

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.md'), 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='schemaflow',
    author='Jorge C. Leitao',
    version='0.2.0',
    description='A package to write schema-aware data pipelines',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jorgecarleitao/schemaflow",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import unittest

import numpy as np

from schemaflow import pipe, types, exceptions


class Pipe(pipe.Pipe):
    requirements = {'sklearn'}

    # variables required by fit (supervised learning)
    fit_requires = {
        # (arbitrary items, arbitrary features)
        'x': types.Array(np.float64, shape=(None, None)),
        'y': types.List(float)
    }

    transform_requires = {
        'x': types.List(float)
    }

    # parameter passed to fit()
    fit_parameters = {
        'alpha': float
    }

    # parameter assigned in fit()
    fitted_parameters = {
        'model': object
    }

    # type and key of transform
    transform_modifies = {
        'model': object
    }

    def fit(self, data, parameters=None):
        import sklearn.linear_model
        self['model'] = sklearn.linear_model.Lasso(parameters['alpha'])

        self['model'].fit(data['x'], data['y'])

    def transform(self, data):
        data['model'] = self['model']
        return data


class TestRequirements(unittest.TestCase):
    def test_basic(self):
        self.assertEqual(Pipe().check_requirements, [])


class TestPipe(unittest.TestCase):

    def test_fit_transform(self):
        p = Pipe()

        p.fit({'x': [[1.0], [2.0]], 'y': [1.0, 1.0]}, {'alpha': 0.1})
        data = p.transform({'df1': [[1.0], [2.0]]})

        self.assertEqual(data['model'].__class__.__name__, 'Lasso')

    def test_check_fit_parameters(self):
        p = Pipe()
        good_data = {'x': np.array([[1.0], [2.0]]), 'y': [1.0, 1.0]}

        errors = p.check_fit(good_data, {'alph': 0.1})
        self.assertEqual(len(errors), 1)
        self.assertEqual(type(errors[0]), exceptions.WrongParameter)

        # use default parameter
        errors = p.check_fit(good_data)
        self.assertEqual(len(errors), 1)
        self.assertEqual(type(errors[0]), exceptions.WrongParameter)

        with self.assertRaises(exceptions.WrongParameter) as e:
            p.check_fit(good_data, {'alph': 0.1}, raise_=True)
        self.assertIn('in fit', str(e.exception))

        errors = p.check_fit(good_data, {'alpha': 'a'})
        self.assertEqual(len(errors), 1)
        self.assertEqual(type(errors[0]), exceptions.WrongType)

        with self.assertRaises(exceptions.WrongType) as e:
            p.check_fit(good_data, {'alpha': 'a'}, raise_=True)
        self.assertIn('in parameter \'alpha\' of fit', str(e.exception))

    def test_check_fit(self):
        p = Pipe()

        bad_data_schema = {'x': np.array([[1.0], [2.0]]), 'z': [1.0, 1.0]}
        bad_data_type = {'x': 1, 'y': [1.0, 1.0]}
        good_data = {'x': np.array([[1.0], [2.0]]), 'y': [1.0, 1.0]}

        errors = p.check_fit(good_data, {'alpha': 0.1})
        self.assertEqual(len(errors), 0)

        errors = p.check_fit(bad_data_schema, {'alpha': 0.1})
        self.assertEqual(len(errors), 1)
        self.assertEqual(type(errors[0]), exceptions.WrongSchema)

        with self.assertRaises(exceptions.WrongSchema) as e:
            p.check_fit(bad_data_schema, {'alpha': 0.1}, raise_=True)
        self.assertIn('in fit', str(e.exception))

        errors = p.check_fit(bad_data_type, {'alpha': 0.1})
        self.assertEqual(len(errors), 1)
        self.assertEqual(type(errors[0]), exceptions.WrongType)

    def test_check_transform(self):
        p = Pipe()

        bad_data_schema = {'x1': np.array([[1.0], [2.0]])}
        bad_data_type = {'x': 1}
        good_data = {'x': [1.0]}

        errors = p.check_transform(good_data)
        self.assertEqual(len(errors), 0)

        errors = p.check_transform(bad_data_schema)
        self.assertEqual(len(errors), 1)
        self.assertEqual(type(errors[0]), exceptions.WrongSchema)

        with self.assertRaises(exceptions.WrongSchema) as e:
            p.check_transform(bad_data_schema, raise_=True)
        self.assertIn('in transform', str(e.exception))

        errors = p.check_transform(bad_data_type)
        self.assertEqual(len(errors), 1)
        self.assertEqual(type(errors[0]), exceptions.WrongType)

        with self.assertRaises(exceptions.WrongType) as e:
            p.check_transform(bad_data_type, raise_=True)
        self.assertIn('in argument \'x\' of transform', str(e.exception))

    def test_not_fitted(self):
        p = Pipe()
        with self.assertRaises(exceptions.NotFittedError) as e:
            p['model']
        self.assertIn('model', str(e.exception))

    def test_transform_schema(self):
        p = Pipe()

        self.assertEqual(p.transform_schema({'x': types.List(float)}), {
            'x': types.List(float),
            'model': object})

        with self.assertRaises(exceptions.WrongSchema) as e:
            p.transform_schema({'y': types.List(float)})
        self.assertIn('in transform', str(e.exception))



import schemaflow.types


class Operation:
    """
    Declares a generic operation to the schema, used in :attr:`~schemaflow.pipe.Pipe.transform_modifies`.
    """
    def transform(self, key, schema):
        return schema


class Set(Operation):
    """
    Declares an operation that sets a new schema key.

    Not using an operation to define
    """
    def __init__(self, value_type):
        if not isinstance(value_type, schemaflow.types._LiteralType):
            value_type = schemaflow.types._LiteralType(value_type)
        self.value_type = value_type

    def transform(self, key, schema):
        schema[key] = self.value_type
        return schema

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.value_type)


class Drop(Operation):
    """
    Declares an operation that drops an existing schema key.
    """
    def transform(self, key, schema):
        return {k: value for k, value in schema.items() if k != key}

    def __repr__(self):
        return '%s()' % self.__class__.__name__


class ModifyDataFrame(Operation):
    """
    Declares an operation that modifies a schema from a DataFrame.
    """
    def __init__(self, ops):
        self.ops = ops

    def transform(self, key, schema):
        instance = schema.get(key, schemaflow.types._DataFrame({}))

        for column, op in self.ops.items():
            instance.schema = op.transform(column, instance.schema)
        return schema

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.ops)
